@@ -3,6 +3,7 @@ import sys
 from vipy.util import try_import, islist, tolist, tempdir, remkdir, chunklistbysize, listpkl, filetail, filebase, tempdir
 from itertools import repeat
 import dill
+dill.extend(False)  # https://github.com/uqfoundation/dill/issues/383
 try_import('dask', 'dask distributed torch')
 from dask.distributed import as_completed, wait
 try_import('torch', 'torch');  import torch
@@ -18,7 +19,7 @@ import vipy.globals
 import hashlib
 import uuid
 import shutil
-
+dill.extend(True)  # https://github.com/uqfoundation/dill/issues/383
 
 
 class Dask(object):
@@ -29,9 +30,6 @@ class Dask(object):
             assert num_processes is None, "Cannot specify both num_gpus and num_processes"
             num_processes = num_gpus   # coercing
 
-        dask.config.set(distributed__comm__timeouts__tcp="60s")
-        dask.config.set(distributed__comm__timeouts__connect="60s")        
-
         self._num_processes = num_processes
 
         if address is not None:
@@ -48,9 +46,16 @@ class Dask(object):
                                   n_workers=num_processes, 
                                   env={'VIPY_BACKEND':'Agg',  # headless 
                                        'PYTHONOPATH':os.environ['PYTHONPATH'] if 'PYTHONPATH' in os.environ else '',
-                                       'PATH':os.environ['PATH'] if 'PATH' in os.environ else ''},
+                                       'PATH':os.environ['PATH'] if 'PATH' in os.environ else '',
+                                       #'DASK_DISTRIBUTED__COMM__TIMEOUTS__TCP':"30s",
+                                       #'DASK_DISTRIBUTED__COMM__TIMEOUTS__CONNECT':"30s",
+                                       #'DASK_DISTRIBUTED__SCHEDULER__WORK_STEALING':'false',
+                                       #'DASK_DISTRIBUTED__ADMIN__TICK__LIMIT':"30s",
+                                       #'DASK_DISTRIBUTED__ADMIN__TICK__INTERVAL':"2s",
+                                       #'DASK_DISTRIBUTED__DEPLOY__LOST_WORKER_TIMEOUT':"30s"
+                                   },
                                   direct_to_workers=True,
-                                  silence_logs=(False if isdebug() else 30) if not verbose else 40,  # logging.WARN or logging.ERROR or logging.INFO
+                                  silence_logs=20 if verbose else 40,  # logging.WARN=30 or logging.ERROR=40 or logging.INFO=20
                                   local_directory=tempfile.mkdtemp())
 
         self._num_gpus = num_gpus
@@ -80,13 +85,13 @@ class Dask(object):
         webbrowser.open(self._client.dashboard_link) if len(self._client.dashboard_link)>0 else None
     
     def num_processes(self):
-        return self._num_processes
+        return len(self._client.nthreads()) if self._client is not None else 0
 
     def shutdown(self):
         self._client.close()
         self._num_processes = None
         self._num_gpus = None
-        
+        vipy.globals.GLOBAL['DASK_CLIENT'] = None        
         return self
 
     def client(self):
@@ -174,29 +179,13 @@ class Batch(Checkpoint):
 
     """    
              
-    def __init__(self, objlist, strict=True, as_completed=False, checkpoint=False, checkpointdir=None, checkpointfrac=0.1):
+    def __init__(self, objlist, strict=True, as_completed=False, checkpoint=False, checkpointdir=None, checkpointfrac=0.1, warnme=True, minscatter=None):
         """Create a batch of homogeneous vipy.image objects from an iterable that can be operated on with a single parallel function call
         """
         assert isinstance(objlist, list), "Input must be a list"
         self._batchtype = type(objlist[0]) if len(objlist)>0 else type(None)
         assert all([isinstance(im, self._batchtype) for im in objlist]), "Invalid input - Must be homogeneous list of the same type"                
-        self._objlist = objlist
-
-        # This has been deprecated - Set up globally
-        #if n_processes is not None and ngpu is not None:
-        #    assert n_processes == ngpu, "Number of processes must be equal to the number of GPUs"
-        #elif n_processes is not None:
-        #    n_processes = ngpu if (ngpu is not None) and isinstance(ngpu, int) and ngpu > 0 else n_processes
-        #    n_processes = vipy.globals.max_workers() if n_processes is None else n_processes      
-        #    if vipy.globals.dask() is None or n_processes != vipy.globals.dask().num_processes():
-        #        assert n_processes is not None, "set vipy.globals.max_workers() or n_processes kwarg"
-        #        vipy.globals.dask(num_processes=n_processes, dashboard=dashboard)
-        #elif ngpu is not None:
-        #    n_processes = ngpu
-        #    if vipy.globals.dask() is None or n_processes != vipy.globals.dask().num_processes():
-        #        vipy.globals.dask(num_processes=n_processes, dashboard=dashboard)
-        #else:
-        #    assert vipy.globals.dask() is not None, "Create a global dask scheduler using vipy.globals.dask(num_processes=n) for a given number of workers, or create Batch() with n_processes>0"            
+        self._objlist = [(k,o) for (k,o) in enumerate(objlist)]  # ordered
 
         assert checkpointfrac > 0.0 and checkpointfrac <= 1.0, "Invalid checkpoint fraction"
         self._checkpointsize = max(1, int(len(objlist) * checkpointfrac))
@@ -205,11 +194,9 @@ class Batch(Checkpoint):
         if checkpoint:
             as_completed = True  # force self._as_completed=True
 
-        if vipy.globals.dask() is None:
-            warnings.warn('vipy.batch.Batch() is not set to use parallelism.  This is set using:\n>>> vipy.globals.dask(num_processes=n) for multi-processing\n>>> vipy.globals.dask(num_gpus=m) for multi-gpu\n>>> vipy.globals.dask(pct=0.8) for multiprocessing that uses a percentage of the current system resources\n>>> vipy.globals.dask(address="SCHEDULER:PORT") which connects to a Dask distributed scheduler.\n>>> vipy.globals.noparallel() to completely disable all parallelism.')
+        if vipy.globals.dask() is None and warnme:
+            print('[vipy.batch.Batch]: vipy.batch.Batch() is not set to use parallelism.  This is set using:\n    >>> vipy.globals.parallel(n) for multi-processing with n processes\n    >>> vipy.globals.parallel(pct=0.8) for multiprocessing that uses a percentage of the current system resources\n    >>> vipy.globals.dask(address="SCHEDULER:PORT") which connects to a Dask distributed scheduler.\n    >>> vipy.globals.noparallel() to completely disable all parallelism.')
 
-        #self._client = vipy.globals.dask().client() if vipy.globals.dask() is not None else None
-    
         # FIXME: this needs to go into Dask()
         #self._ngpu = ngpu
         #if ngpu is not None:
@@ -218,7 +205,7 @@ class Batch(Checkpoint):
 
         self._strict = strict
         self._as_completed = as_completed  # this may introduce instabilities for large complex objects, use with caution
-
+        self._minscatter = minscatter
 
     def __enter__(self):
         return self
@@ -232,19 +219,9 @@ class Batch(Checkpoint):
     def __repr__(self):        
         return str('<vipy.batch: type=%s, len=%d, client=%s>' % (str(self._batchtype), len(self), str(vipy.globals.dask())))
 
-    #def ngpu(self):
-    #    return self._ngpu
-
     def _client(self):
         return vipy.globals.dask()._client if vipy.globals.dask() is not None else None
 
-    #def shutdown(self):
-    #    vipy.globals.dask().shutdown()
-
-    #def n_processes(self):
-    #    return len(self.info()['workers'])
-
-
     def _batch_wait(self, futures):
         self._archive_checkpoint()._flush_checkpoint()
         k_checkpoint = 0
@@ -253,7 +230,7 @@ class Batch(Checkpoint):
             for (k, batch) in enumerate(as_completed(futures, with_results=True, raise_errors=False).batches()):
                 for (future, result) in batch:
                     if future.status != 'error':
-                        results.append(result)  # not order preserving
+                        results.append(result)  # not order preserving, will restore order in result()
                     else:
                         if self._strict:
                             typ, exc, tb = result
@@ -262,6 +239,8 @@ class Batch(Checkpoint):
                             print('[vipy.batch]: future %s failed with error "%s" - SKIPPING' % (str(future), str(result)))
                         results.append(None)
                     k_checkpoint = k_checkpoint + 1
+                    
+                    del future, result  # distributed memory cleanup
 
                 # Save intermediate results
                 if self._checkpoint and (k_checkpoint > self._checkpointsize):
@@ -270,6 +249,9 @@ class Batch(Checkpoint):
                     vipy.util.save(results[-k_checkpoint:], pklfile)
                     k_checkpoint = 0
 
+                # Distributed memory cleanup
+                del batch
+  
             return results
 
         except KeyboardInterrupt:
@@ -284,8 +266,7 @@ class Batch(Checkpoint):
     def _wait(self, futures):
         assert islist(futures) and all([hasattr(f, 'result') for f in futures])
         if self._as_completed:
-            return self._batch_wait(futures)
-        
+            return self._batch_wait(futures)        
         try:
             results = []            
             wait(futures)
@@ -314,92 +295,51 @@ class Batch(Checkpoint):
         return self.checkpoint() if self._checkpoint else None
 
     def result(self):
-        """Return the result of the batch processing"""
-        return self._objlist
+        """Return the result of the batch processing, ordered"""
+        objlist = {int(v[0]):v[1] for v in self._objlist if v is not None}
+        return [objlist[k] if k in objlist else None for k in range(len(self._objlist))]  # restore order
 
     def __iter__(self):
-        for im in self._objlist:
-            yield im
+        for x in self.result():
+            yield x
             
-    def product(self, f_lambda, args):
-        """Cartesian product of args and batch.  Use this with extreme caution, as the memory requirements may be high."""
-        assert self._client() is not None, "Invalid Batch() backend"
-        c = self._client()
-        objlist = c.scatter(self._objlist)        
-        self._objlist = self._wait([c.submit(f_lambda, im, *a) for im in objlist for a in args])
-        return self
-
     def map(self, f_lambda, args=None):
         """Run the lambda function on each of the elements of the batch and return the batch object.
         
-        If args is provided, then this is a unique argument for the lambda function for each of the elements in the batch, or is broadcastable.
-        
         >>> iml = [vipy.image.RandomScene(512,512) for k in range(0,1000)]   
-        >>> imb = vipy.image.Batch(iml, n_processes=4) 
-        >>> imb.map(lambda im,f: im.saveas(f), args=[('/tmp/out%d.jpg'%k,) for k in range(0,1000)])  
-        >>> imb.map(lambda im: im.rgb())  # this is equivalent to imb.rgb()
+        >>> imb = vipy.image.Batch(iml) 
+        >>> imb.map(lambda im: im.rgb())  
 
         The lambda function f_lambda must not include closures.  If it does, construct the batch with tuples (obj,prms) or with default parameter capture:
         >>> f = lambda x, prm1=1, prm2=2: x+prm1+prm2
 
         """
         c = self._client()
+
         if c is None:
-            self._objlist = [f_lambda(o) for o in self._objlist]  # no parallelism
-        elif args is not None:
-            if len(self._objlist) > 1:
-                assert islist(args) and len(list(args)) == len(self._objlist), "args must be a list of arguments of length %d, one for each element in batch" % len(self._objlist)
-                objlist = c.scatter(self._objlist)
-                self._objlist = self._wait([c.submit(f_lambda, im, *a) for (im, a) in zip(objlist, args)])
-            else:
-                assert islist(args), "args must be a list"
-                obj = c.scatter(self._objlist[0], broadcast=True)
-                self._objlist = self._wait([c.submit(f_lambda, obj, *a) for a in args])
+            f_lambda_ordered = lambda x,f=f_lambda: (x[0], f(x[1]))                         
+            self._objlist = [f_lambda_ordered(o) for o in self._objlist]  # no parallelism
         else:
-            self._objlist = self._wait(c.map(f_lambda, self._objlist))
+            f_lambda_ordered = lambda x,f=f_lambda: (x[0], f(x[1]))            
+            objlist = c.scatter(self._objlist) if (self._minscatter is not None and len(self._objlist) > self._minscatter) else self._objlist
+            self._objlist = self._wait(c.map(f_lambda_ordered, objlist))
         return self
 
     def filter(self, f_lambda):
         """Run the lambda function on each of the elements of the batch and filter based on the provided lambda keeping those elements that return true 
         """
         c = self._client()
+        f_lambda_ordered = lambda x,f=f_lambda: (x[0], f(x[1])) 
+
         if c is None:
-            self._objlist = [o for o in self._objlist if f_lambda(o)]  # no parallelism
+            self._objlist = [o for o in self._objlist if f_lambda_ordered(o)[1]]  # no parallelism
         else:
             objlist = self._objlist  # original list
-            is_filtered = self._wait(c.map(f_lambda, c.scatter(self._objlist)))  # distributed filter (replaces self._objlist)
-            self._objlist = [obj for (f, obj) in zip(is_filtered, objlist) if f is True]  # keep only elements that filter true
+            is_filtered = self._wait(c.map(f_lambda_ordered, c.scatter(self._objlist)))  # distributed filter (replaces self._objlist)
+            self._objlist = [obj for (f, obj) in zip(is_filtered, objlist) if f[1] is True]  # keep only elements that filter true
         return self
         
-    #def torch(self):
-    #    """Convert the batch of N HxWxC images to a NxCxHxW torch tensor"""
-    #    return torch.cat(self.map(lambda im: im.torch()).result())
-
-    #def numpy(self):
-    #    """Convert the batch of N HxWxC images to a NxCxHxW torch tensor"""
-    #    return np.stack(self.map(lambda im: im.numpy()).result())
-    
-    #def torchmap(self, f, net):
-    #    """Apply lambda function f prepare data, and pass resulting tensor to data parallel to torch network"""
-    #    assert self.__dict__['_client'] is not None, "Batch() must be reconsructed after shutdown"        
-    #
-    #    c = self.__dict__['_client']       
-    #    deviceid = 'cuda' if torch.cuda.is_available and self.ngpu() > 0 else 'cpu'
-    #    device = torch.device(deviceid)
-    #
-    #    modeldist = torch.nn.DataParallel(net)
-    #    modeldist = modeldist.to(device)
-    #    
-    #    return [modeldist(t.to(device)) for t in as_completed([c.submit(f, im) for im in self._objlist])]
-        
-    #def chunkmap(self, f, obj, batchsize):
-    #    c = self._client()
-    #    assert c is not None, "Invalid backend"
-    #    objdist = c.scatter(obj, broadcast=True)        
-    #    self._objlist = self._wait([c.submit(f, objdist, imb) for imb in chunklistbysize(self._objlist, batchsize)])
-    #    return self
-
-    def scattermap(self, f, obj):
+    def scattermap(self, f_lambda, obj):
         """Scatter obj to all workers, and apply lambda function f(obj, im) to each element in batch
         
            Usage: 
@@ -415,11 +355,14 @@ class Batch(Checkpoint):
 
         """
         c = self._client()
+        f_lambda_ordered = lambda net,x,f=f_lambda: (x[0], f(net,x[1])) 
+
         if c is None:
-            self._objlist = [f(obj, o) for o in self._objlist]  # no parallelism
+            self._objlist = [f_lambda_ordered(obj, o) for o in self._objlist]  # no parallelism
         else:
             objdist = c.scatter(obj, broadcast=True)        
-            self._objlist = self._wait([c.submit(f, objdist, im) for im in self._objlist])
+            objlist = c.scatter(self._objlist) if len(self._objlist) > self._minscatter else self._objlist
+            self._objlist = self._wait([c.submit(f_lambda_ordered, objdist, im) for im in objlist])
         return self
 
 
